body {
    background-color: white;
    font-family: Verdana, sans-serif;
    color: #888888;
}

.index-body {
    background-color: white;
    font-family: Verdana, sans-serif;
    color: #888888;
}

table {
    background: #cccccc;
    border-left: 1px outset #b2b2b2;
    border-top: 1px outset #b2b2b2;
    border-right: 1px outset #5b5b5b;
    border-bottom: 1px outset #5b5b5b;
    border-spacing: 5px;
}

th, td {
    border: 1px inset #888888;
    background-color: #777777;
    padding: 0;
}

.cgi-header-table {
    background: #cccccc;
    border: 1px black;
    border-left: 1px outset #b2b2b2;
    border-top: 1px outset #b2b2b2;
    border-right: 1px outset #5b5b5b;
    border-bottom: 1px outset #5b5b5b;
    border-spacing: 5px;
    font-size: 24px;
}



.startpage-header-table {
    border: 0;
    background: transparent;
}

.table-module {
    background: #cccccc;
    border: 1px black;
    border-left: 1px outset #b2b2b2;
    border-top: 1px outset #b2b2b2;
    border-right: 1px outset #5b5b5b;
    border-bottom: 1px outset #5b5b5b;
    border-spacing: 5px;
}

.td-title {
    background-color: #777777;
    font-family: Verdana, sans-serif;
    color: #cccc00;
}

.text-title-multihost {
    font-family: Verdana, sans-serif;
    color: black;
}

.td-valign-top {
    vertical-align: top;
}

.td-title-host {
    font-family: Verdana, sans-serif;
    background-color: white;
    color: black;
}

.td-note {
    background-color: #777777;
    font-family: Verdana, sans-serif;
    color: #cccc00;
}

.text-title-date {
    font-family: Verdana, sans-serif;
    color: #888888;
}
 
.text-title-multihost-one {
    background-color: #777777;
    font-family: Verdana, sans-serif;
    color: black;
}

.text-title {
    font-family: Verdana, sans-serif;
    color: #cccc00;
}

.text-copyright {
    font-size: 10px;
    color: black;
}

.text-size {
    font-size: 5;
}

.history01 {
    color: #777777;
    position: fixed;
    left: 1em;
    font-size: 32px;
    letter-spacing: -1px;
}
